import argparse
from pathlib import Path
from typing import List, Optional

import numpy as np

from mohou.dataset import AutoRegressiveDatasetConfig
from mohou.encoder import ImageEncoder
from mohou.encoding_rule import EncodingRule
from mohou.file import get_project_path
from mohou.model.lstm import LSTM, LSTMConfig
from mohou.script_utils import create_default_logger, train_lstm
from mohou.setting import setting
from mohou.trainer import TrainCache, TrainConfig
from mohou.types import EpisodeBundle


def create_default_image_context_list(
    project_path: Path, bundle: Optional[EpisodeBundle] = None
) -> List[np.ndarray]:
    if bundle is None:
        bundle = EpisodeBundle.load(project_path)
    image_encoder = ImageEncoder.create_default(project_path)

    context_list = []
    for episode in bundle:
        seq = episode.get_sequence_by_type(image_encoder.elem_type)
        context = image_encoder.forward(seq[0])
        context_list.append(context)

    return context_list


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-pn", type=str, default=setting.primary_project_name, help="project name")
    parser.add_argument("-pp", type=str, help="project path. preferred over pn.")
    parser.add_argument("-n", type=int, default=3000, help="iteration number")
    parser.add_argument("-aug", type=int, default=2, help="number of augmentation X")
    parser.add_argument("-hidden", type=int, default=200, help="number of hidden state of lstm")
    parser.add_argument(
        "-window",
        type=int,
        help="window size of episode cutting. If None, episode will not be cut.",
    )
    parser.add_argument("-layer", type=int, default=4, help="number of layers of lstm")
    parser.add_argument("-cov-scale", type=float, default=0.1, help="covariance scale in aug")
    parser.add_argument(
        "-avbias-std", type=float, default=0.0, help="std of angle vector calibration bias std"
    )
    parser.add_argument(
        "-valid-ratio", type=float, default=0.1, help="split rate for validation dataset"
    )
    parser.add_argument("--variational", action="store_true", help="variational")
    parser.add_argument("--warm", action="store_true", help="warm start")
    parser.add_argument(
        "--use_image_context", action="store_true", help="initial image as context input"
    )
    args = parser.parse_args()

    project_name: str = args.pn
    project_path_str: Optional[str] = args.pp
    n_epoch: int = args.n
    n_aug: int = args.aug
    n_hidden: int = args.hidden
    n_layer: int = args.layer
    window_size: Optional[int] = args.window
    cov_scale: float = args.cov_scale
    av_calibration_bias_std: float = args.avbias_std
    valid_ratio: float = args.valid_ratio
    warm_start: bool = args.warm
    variational_lstm: bool = args.variational
    use_image_context: bool = args.use_image_context

    if project_path_str is None:
        project_path = get_project_path(project_name)
    else:
        project_path = Path(project_path_str)

    logger = create_default_logger(project_path, "lstm")  # noqa

    encoding_rule = EncodingRule.create_default(project_path)
    model_config = LSTMConfig(
        encoding_rule.dimension,
        n_hidden=n_hidden,
        n_layer=n_layer,
        variational=variational_lstm,
        type_wise_loss=True,
        type_bound_table=encoding_rule.type_bound_table,
    )
    dataset_config = AutoRegressiveDatasetConfig(
        n_aug=n_aug,
        cov_scale=cov_scale,
        window_size=window_size,
        av_calibration_bias_std=av_calibration_bias_std,
    )
    train_config = TrainConfig(n_epoch=n_epoch, valid_data_ratio=valid_ratio)

    context_list = None
    if use_image_context:
        context_list = create_default_image_context_list(project_path)
        model_config.n_static_context = len(context_list[0])

    if warm_start:
        tcache_pretrained: Optional[TrainCache[LSTM]] = TrainCache.load_latest(project_path, LSTM)
    else:
        tcache_pretrained = None

    train_lstm(
        project_path,
        encoding_rule,
        model_config,
        dataset_config,
        train_config,
        tcache_pretrained=tcache_pretrained,
        context_list=context_list,
    )



from dataclasses import dataclass
from typing import List, Optional, Tuple, Type, TypeVar, Union

import torch
import torch.nn as nn

from mohou.model.common import LossDict, ModelBase, ModelConfigBase
from mohou.model.lstm import LSTM, LSTMConfig


def build_linear_layers(
    n_input: int, n_output: int, n_hidden: int, n_layer: int, activation: Optional[str]
) -> List[nn.Module]:

    if activation is not None:
        assert activation in ("relu", "sigmoid", "tanh")

    AT: Optional[Type[nn.Module]] = None
    if activation == "relu":
        AT = nn.ReLU
    elif activation == "sigmoid":
        AT = nn.Sigmoid
    elif activation == "tanh":
        AT = nn.Tanh

    layers: List[nn.Module] = []
    input_layer = nn.Linear(n_input, n_hidden)
    layers.append(input_layer)
    if AT is not None:
        layers.append(AT())

    for _ in range(n_layer):
        middle_layer = nn.Linear(n_hidden, n_hidden)
        layers.append(middle_layer)
        if AT is not None:
            layers.append(AT())

    output_layer = nn.Linear(n_hidden, n_output)
    layers.append(output_layer)
    return layers


@dataclass
class MarkoveModelConfig(ModelConfigBase):
    n_input: int
    n_output: int
    n_hidden: int = 200
    n_layer: int = 4
    activation: Optional[str] = None  # TODO(HiroIshida): consider replace it with enum

    def __post_init__(self):
        if self.activation is not None:
            assert self.activation in ("relu", "sigmoid", "tanh")


class MarkovPredictionModel(ModelBase):
    """f(s, a) -> (s, a) type model"""

    def _setup_from_config(self, config: MarkoveModelConfig) -> None:
        config.n_input
        layers = build_linear_layers(
            n_input=config.n_input,
            n_output=config.n_output,
            n_hidden=config.n_hidden,
            n_layer=config.n_layer,
            activation=config.activation,
        )
        self.layer = nn.Sequential(*layers)

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> LossDict:
        # NOTE: episode index and static context is not supported in this model
        # suppose using AutoRegressiveDataset
        _, seq_sample, _ = sample
        X0 = seq_sample[:, :-1]
        X1 = seq_sample[:, 1:]

        X1_pred = self.layer(X0)
        loss = nn.MSELoss()(X1, X1_pred)
        return LossDict({"prediction": loss})

    def forward(self, seq_sample: torch.Tensor) -> torch.Tensor:
        n_batch, n_seq_len, n_dim = seq_sample.shape
        sample_pre = seq_sample.reshape(-1, n_dim)
        sample_post = self.layer(sample_pre)
        seq_sample_post = sample_post.reshape(seq_sample.shape)
        return seq_sample_post


class ControlModel(ModelBase):
    """f(s, a) -> s type model"""

    layer: nn.Sequential

    def _setup_from_config(self, config: MarkoveModelConfig) -> None:
        config.n_input
        layers = build_linear_layers(
            n_input=config.n_input,
            n_output=config.n_output,
            n_hidden=config.n_hidden,
            n_layer=config.n_layer,
            activation=config.activation,
        )
        self.layer = nn.Sequential(*layers)

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> LossDict:
        inp_ctrl_sample, inp_obs_sample, out_obs_sample = sample
        inp_sample = torch.concat((inp_ctrl_sample, inp_obs_sample), dim=1)
        out_obs = self.layer(inp_sample)
        loss = nn.MSELoss()(out_obs_sample, out_obs)
        return LossDict({"prediction": loss})


@dataclass
class ProportionalModelConfig(ModelConfigBase):
    n_input: int
    n_bottleneck: int = 6
    n_layer: int = 2


class ProportionalModel(ModelBase[ProportionalModelConfig]):
    # This model is highly experimental. Maybe deleted without any notification.
    encoder: nn.Module
    decoder: nn.Module
    propagator: nn.Module
    # p_value: Parameter
    p_value: nn.Module

    def _setup_from_config(self, config: ProportionalModelConfig) -> None:
        layers = build_linear_layers(
            config.n_input, config.n_bottleneck, 100, config.n_layer, activation="tanh"
        )
        self.encoder = nn.Sequential(*layers)

        layers = build_linear_layers(
            config.n_bottleneck, config.n_input, 100, config.n_layer, activation="tanh"
        )
        self.decoder = nn.Sequential(*layers)

        # param = Parameter(torch.ones(1) * 0.0)
        # self.register_parameter("kp", param)
        layers = build_linear_layers(config.n_bottleneck, 1, config.n_bottleneck * 2, 2, "tanh")
        layers.append(nn.Tanh())
        self.p_value = nn.Sequential(*layers)

    def get_p_value(self, bottoleneck_tensor: torch.Tensor) -> torch.Tensor:
        return 0.5 * (self.p_value(bottoleneck_tensor) + 0.5)

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> LossDict:
        # NOTE: episode index and static context is not supported in this model
        _, seq_sample, _ = sample

        n_batch, n_seq_len, n_dim = seq_sample.shape
        X0 = seq_sample.reshape(-1, n_dim)
        Z0 = self.encoder(X0)
        Z0_reshaped = Z0.reshape(n_batch, n_seq_len, self.config.n_bottleneck)

        # compute reconstruction loss
        f_loss = nn.MSELoss()
        X0_hat = self.decoder(Z0)
        reconstruction_loss = f_loss(X0_hat, X0)

        # compute prediction loss
        prediction_loss_list: List[torch.Tensor] = []
        n_window_desired = 10
        n_window_max = n_seq_len - 1
        n_window = min(n_window_desired, n_window_max)

        Z_prop_est = Z0
        for window in range(1, n_window + 1):
            KP = self.get_p_value(Z_prop_est)
            # print("min: {}, max {}".format(torch.min(KP), torch.max(KP)))
            Z_prop_est = Z_prop_est * (1 - KP)

            Z_prop_est_reshaped = Z_prop_est.reshape(n_batch, n_seq_len, self.config.n_bottleneck)
            Z_prop_est_reshaped_cut = Z_prop_est_reshaped[:, :-window, :]
            Z_prop_reshaped_cut = Z0_reshaped[:, window:, :]
            partial_loss = nn.MSELoss()(Z_prop_est_reshaped_cut, Z_prop_reshaped_cut)
            prediction_loss_list.append(partial_loss)

        prediction_loss = torch.sum(torch.stack(prediction_loss_list)) / len(prediction_loss_list)
        return LossDict({"reconstruction": reconstruction_loss, "prediction": prediction_loss})

    def forward(self, seq_sample: torch.Tensor) -> torch.Tensor:
        n_batch, n_seq_len, n_dim = seq_sample.shape
        sample_pre = seq_sample.reshape(-1, n_dim)
        z = self.encoder(sample_pre)
        p_value = self.get_p_value(z)
        z_post = (1.0 - p_value) * z
        sample_post = self.decoder(z_post)
        return sample_post


MarkovModelT = TypeVar("MarkovModelT", bound=Union[MarkovPredictionModel, ProportionalModel])


@dataclass
class DisentangleLSTMConfig(ModelConfigBase):
    n_state_with_flag: int
    n_bottleneck: int = 10
    n_fc_layer: int = 3
    n_fc_hidden: int = 100
    n_lstm_hidden: int = 200
    n_lstm_layer: int = 1


class DisentangleLSTM(ModelBase[DisentangleLSTMConfig]):
    encoder: nn.Module
    decoder: nn.Module
    lstm: LSTM

    def _setup_from_config(self, config: DisentangleLSTMConfig) -> None:
        layers = build_linear_layers(
            config.n_state_with_flag,
            config.n_bottleneck,
            config.n_fc_hidden,
            config.n_fc_layer,
            activation="tanh",
        )
        self.encoder = nn.Sequential(*layers)

        layers = build_linear_layers(
            config.n_bottleneck,
            config.n_state_with_flag,
            config.n_fc_hidden,
            config.n_fc_layer,
            activation="tanh",
        )
        self.decoder = nn.Sequential(*layers)

        conf = LSTMConfig(
            config.n_bottleneck, 0, config.n_lstm_hidden, config.n_lstm_layer, n_output_layer=1
        )
        self.lstm = LSTM(conf)

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> LossDict:
        # NOTE: episode index and static context is not supported in this model
        _, X, _ = sample

        n_batch, n_seq_len, n_dim = X.shape
        X0 = X[:, :-1, :]
        X1 = X[:, 1:, :]

        X0_reshaped = X0.reshape(-1, n_dim)
        Z0_reshaped = self.encoder(X0_reshaped)
        X0_reconst_reshaped = self.decoder(Z0_reshaped)
        X0_reconst = X0_reconst_reshaped.reshape(n_batch, n_seq_len - 1, n_dim)

        Z0 = Z0_reshaped.reshape(n_batch, n_seq_len - 1, self.config.n_bottleneck)

        empty_context = torch.empty((n_batch, 0)).to(Z0.device)
        Z1_est, _ = self.lstm.forward(Z0, empty_context)
        Z1_est_reshaped = Z1_est.reshape(-1, self.config.n_bottleneck)
        X1_est_reshaped = self.decoder(Z1_est_reshaped)
        X1_est = X1_est_reshaped.reshape(n_batch, n_seq_len - 1, n_dim)

        reconstruction_loss = nn.MSELoss()(X0_reconst, X0)
        prediction_loss = nn.MSELoss()(X1_est, X1)

        return LossDict({"reconstruction": reconstruction_loss, "prediction": prediction_loss})

    def forward(self, X0: torch.Tensor) -> torch.Tensor:
        n_batch, n_seq_len, n_dim = X0.shape

        X0_reshaped = X0.reshape(-1, n_dim)

        Z0_reshaped = self.encoder(X0_reshaped)

        Z0 = Z0_reshaped.reshape(n_batch, n_seq_len, self.config.n_bottleneck)

        empty_context = torch.empty((n_batch, 0)).to(Z0.device)
        Z1_est, _ = self.lstm.forward(Z0, empty_context)

        Z1_est_reshaped = Z1_est.reshape(-1, self.config.n_bottleneck)
        X1_est_reshaped = self.decoder(Z1_est_reshaped)
        X1_est = X1_est_reshaped.reshape(n_batch, n_seq_len, n_dim)
        return X1_est



import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import (
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
import torch

from mohou.encoder import EncoderBase, HasAModel, ImageEncoder, VectorIdenticalEncoder
from mohou.types import (
    AngleVector,
    AnotherGripperState,
    CompositeImageBase,
    DepthImage,
    ElementBase,
    ElementDict,
    EpisodeBundle,
    EpisodeData,
    GripperState,
    PrimitiveElementBase,
    RGBDImage,
    RGBImage,
    TerminateFlag,
)
from mohou.utils import assert_equal_with_message, get_bound_list

logger = logging.getLogger(__name__)


ScaleBalancerT = TypeVar("ScaleBalancerT", bound="ScaleBalancerBase")
ArrayT = TypeVar("ArrayT", bound=Union[torch.Tensor, np.ndarray])


class ScaleBalancerBase(ABC):
    @classmethod
    def get_json_file_path(cls, project_path: Path, create_dir: bool = False) -> Path:
        """get the json file path that will be used for dumping and loading
        Note that dumping / loading the scale balancer is still an experimental feature.
        """
        save_dir_path = project_path / "experimental"
        if create_dir:
            save_dir_path.mkdir(exist_ok=True)
        file_name = cls.__name__ + ".json"
        json_file_path = save_dir_path / file_name
        return json_file_path

    @abstractmethod
    def apply(self, arr: ArrayT) -> ArrayT:
        pass

    @abstractmethod
    def inverse_apply(self, arr: ArrayT) -> ArrayT:
        pass

    @abstractmethod
    def dump(self, project_path: Path) -> None:
        pass

    @classmethod
    @abstractmethod
    def load(cls: Type[ScaleBalancerT], project_path: Path) -> ScaleBalancerT:
        pass


@dataclass(frozen=True)
class IdenticalScaleBalancer(ScaleBalancerBase):
    def apply(self, arr: ArrayT) -> ArrayT:
        return arr

    def inverse_apply(self, arr: ArrayT) -> ArrayT:
        return arr

    def dump(self, project_path: Path) -> None:
        json_file_path = self.get_json_file_path(project_path, create_dir=True)
        with json_file_path.open(mode="w") as f:
            json.dump({}, f)

    @classmethod
    def load(cls, project_path: Path) -> "IdenticalScaleBalancer":
        logger.warning("NOTE: This feature may be dropped without any notification")
        json_file_path = cls.get_json_file_path(project_path)
        assert json_file_path.exists()
        with json_file_path.open(mode="r") as f:
            kwargs = json.load(f)
        # because actually this class does not have any members
        assert len(kwargs) == 0
        return cls()

    def __eq__(self, other: object) -> bool:
        keys = self.__dataclass_fields__.keys()  # type: ignore # mypy's bag
        assert set(keys) == set()
        return isinstance(other, IdenticalScaleBalancer)


@dataclass(frozen=True)
class CovarianceBasedScaleBalancer(ScaleBalancerBase):
    dims: List[int]
    means: List[np.ndarray]
    scaled_stds: List[float]

    def __eq__(self, other: object) -> bool:
        keys = self.__dataclass_fields__.keys()  # type: ignore # mypy's bag
        assert set(keys) == {"dims", "means", "scaled_stds"}

        if not isinstance(other, CovarianceBasedScaleBalancer):
            return NotImplemented
        assert type(self) == type(other)

        if self.dims != other.dims:
            return False
        if not np.allclose(np.hstack(self.means), np.hstack(other.means), atol=1e-6):
            return False
        if self.scaled_stds != other.scaled_stds:
            return False

        return True

    def dump(self, project_path: Path):
        json_file_path = self.get_json_file_path(project_path, create_dir=True)
        dic = {
            "dims": self.dims,
            "means": [mean.tolist() for mean in self.means],
            "scaled_stds": self.scaled_stds,
        }
        with json_file_path.open(mode="w") as f:
            json.dump(dic, f)

    @classmethod
    def load(cls, project_path: Path) -> "CovarianceBasedScaleBalancer":
        logger.warning("NOTE: This feature may be dropped without any notification")
        json_file_path = cls.get_json_file_path(project_path)
        assert json_file_path.exists()
        with json_file_path.open(mode="r") as f:
            kwargs = json.load(f)
        means = kwargs["means"]
        kwargs["means"] = [np.array(mean) for mean in means]
        return cls(**kwargs)

    def __post_init__(self):
        for i, dim in enumerate(self.dims):
            assert_equal_with_message(len(self.means), len(self.dims), "len")
            assert_equal_with_message(len(self.scaled_stds), len(self.dims), "len")
            assert_equal_with_message(self.means[i].shape, (dim,), "mean shape of {}".format(i))

    @staticmethod
    def get_bound_list(dims: List[int]) -> List[slice]:
        bound_list = []
        head = 0
        for dim in dims:
            bound_list.append(slice(head, head + dim))
            head += dim
        return bound_list

    @staticmethod
    def is_binary_sequence(partial_feature_seq: np.ndarray):
        return len(set(partial_feature_seq.flatten().tolist())) == 2

    @classmethod
    def from_feature_seqs(cls, feature_seq: np.ndarray, dims: List[int]):
        assert_equal_with_message(feature_seq.ndim, 2, "feature_seq.ndim")
        means = []
        max_stds = []
        for rang in cls.get_bound_list(dims):
            feature_seq_partial = feature_seq[:, rang]
            dim = feature_seq_partial.shape[1]
            if cls.is_binary_sequence(feature_seq_partial):
                # because it's strange to compute covariance for binary sequence
                assert dim == 1, "this restriction maybe removed"
                minn = np.min(feature_seq_partial)
                maxx = np.max(feature_seq_partial)
                cov = np.diag(np.ones(dim))
                mean = np.array([0.5 * (minn + maxx)])
            else:
                mean = np.mean(feature_seq_partial, axis=0)
                cov = np.cov(feature_seq_partial.T)
                if cov.ndim == 0:  # unfortunately, np.cov return 0 dim array instead of 1x1
                    cov = np.expand_dims(cov, axis=0)
                    cov = np.array([[cov.item()]])

            eig_values, _ = np.linalg.eig(cov)
            max_std = np.sqrt(max(eig_values))

            # Note that np.cov(pts.T) outputs some non 0 value even if the input points
            # are all at a single point (degenerated). So, max_eig > 0.0 cannot check
            # the data degeneration. Thus...
            assert max_std > 1e-7

            means.append(mean)
            max_stds.append(max_std)

        scaled_stds = list(np.array(max_stds) / max(max_stds))
        return cls(dims, means, scaled_stds)

    def _apply(self, arr: ArrayT, inverse: bool) -> ArrayT:

        assert arr.ndim in [1, 2]

        dim = len(arr) if arr.ndim == 1 else arr.shape[1]
        assert_equal_with_message(dim, sum(self.dims), "vector total dim")

        mean = np.hstack(self.means)
        std_list = [np.ones(dim) * std for dim, std, in zip(self.dims, self.scaled_stds)]
        std = np.hstack(std_list)

        if isinstance(arr, torch.Tensor):
            mean = torch.from_numpy(mean).float().to(arr.device)  # type: ignore
            std = torch.from_numpy(std).float().to(arr.device)  # type: ignore

        if inverse:
            return (arr * std) + mean  # type: ignore
        else:
            return (arr - mean) / std  # type: ignore

    def apply(self, arr: ArrayT) -> ArrayT:
        return self._apply(arr, False)

    def inverse_apply(self, arr: ArrayT) -> ArrayT:
        return self._apply(arr, True)


class EncodingRuleBase(Mapping[Type[ElementBase], EncoderBase]):
    @abstractmethod
    def apply(self, elem_dict: ElementDict) -> np.ndarray:
        pass

    @abstractmethod
    def inverse_apply(self, vector_processed: np.ndarray) -> ElementDict:
        pass

    @property
    @abstractmethod
    def dimension_list(self) -> List[int]:
        """get dimension list for each encoders output"""

    @property
    def type_bound_table(self) -> Dict[Type[ElementBase], slice]:
        bounds = get_bound_list(self.dimension_list)

        table = {}
        for encoder, bound in zip(self.values(), bounds):
            table[encoder.elem_type] = bound
        return table

    def apply_to_episode_data(self, episode_data: EpisodeData) -> np.ndarray:
        vec_list = [self.apply(episode_data[i]) for i in range(len(episode_data))]
        return np.array(vec_list)

    def apply_to_episode_bundle(self, bundle: EpisodeBundle) -> List[np.ndarray]:
        def elem_types_to_primitive_elem_set(elem_type_list: List[Type[ElementBase]]):
            primitve_elem_type_list = []
            for elem_type in elem_type_list:
                if issubclass(elem_type, PrimitiveElementBase):
                    primitve_elem_type_list.append(elem_type)
                elif issubclass(elem_type, CompositeImageBase):
                    primitve_elem_type_list.extend(elem_type.image_types)
            return set(primitve_elem_type_list)

        bundle_elem_types = elem_types_to_primitive_elem_set(list(bundle.types()))
        required_elem_types = elem_types_to_primitive_elem_set(list(self.keys()))
        assert required_elem_types <= bundle_elem_types

        vector_seq_list = [self.apply_to_episode_data(data) for data in bundle]

        assert vector_seq_list[0].ndim == 2
        return vector_seq_list

    def get_device(self) -> Optional[torch.device]:
        device_list = []
        for encoder in self.values():
            if isinstance(encoder, HasAModel):
                device_list.append(encoder.get_device())
        if len(device_list) == 0:
            return None
        device_set = set(device_list)
        assert len(device_set) == 1, "Do not mix more than 1 devices"
        return device_set.pop()

    def set_device(self, device: torch.device) -> None:
        for encoder in self.values():
            if isinstance(encoder, HasAModel):
                encoder.set_device(device)


_default_encoding_rule_cache: Dict[Tuple, "EncodingRule"] = {}


class EncodingRule(Dict[Type[ElementBase], EncoderBase], EncodingRuleBase):
    scale_balancer: ScaleBalancerBase

    def apply(self, elem_dict: ElementDict) -> np.ndarray:
        vector_list = []
        for elem_type, encoder in self.items():
            vector = encoder.forward(elem_dict[elem_type])
            vector_list.append(vector)
        return self.scale_balancer.apply(np.hstack(vector_list))

    def inverse_apply(self, vector_processed: np.ndarray) -> ElementDict:
        def split_vector(vector: np.ndarray, size_list: List[int]):
            head = 0
            vector_list = []
            for i, size in enumerate(size_list):
                tail = head + size
                vector_list.append(vector[head:tail])
                head = tail
            return vector_list

        vector = self.scale_balancer.inverse_apply(vector_processed)
        size_list = [encoder.output_size for elem_type, encoder in self.items()]
        vector_list = split_vector(vector, size_list)

        elem_dict = ElementDict([])
        for vec, (elem_type, encoder) in zip(vector_list, self.items()):
            elem_dict[elem_type] = encoder.backward(vec)
        return elem_dict

    @property
    def dimension(self) -> int:
        return sum(encoder.output_size for encoder in self.values())

    @property
    def dimension_list(self) -> List[int]:
        return [encoder.output_size for encoder in self.values()]

    @property
    def encode_order(self) -> List[Type[ElementBase]]:
        return list(self.keys())

    def __str__(self) -> str:
        string = "total dim: {}".format(self.dimension)
        for elem_type, encoder in self.items():
            string += "\n{0}: {1}".format(elem_type.__name__, encoder.output_size)
        return string

    @classmethod
    def from_encoders(
        cls,
        encoder_list: Sequence[EncoderBase],
        bundle: Optional[EpisodeBundle] = None,
        scale_balancer: Optional[ScaleBalancerBase] = None,
    ) -> "EncodingRule":
        """Create EncodingRule from encoder_list
        Args:
            encoder_list: list of encoder. Order of the list is important and preserved.
            bundle: If set, ScaleBalancer will created using bundle
            scale_balancer: use this scale balancer if set.

        bundle != None and scale_balancer != None is never accepted
        """
        # NOTE: currently we can load cached balancer. But loading cached entire encoder is,
        # of course the future direction. However, the difficulty mainly lies in the serializing
        # ImageEncoder which contains lambda functions.

        rule: EncodingRule = cls()
        for encoder in encoder_list:
            rule[encoder.elem_type] = encoder
        rule.scale_balancer = IdenticalScaleBalancer()  # tmp

        if scale_balancer is not None:
            assert not bundle
            rule.scale_balancer = scale_balancer

        if bundle is not None:
            assert not scale_balancer
            dims = [encoder.output_size for encoder in rule.values()]
            # compute normalizer and set to encoder
            vector_seqs = rule.apply_to_episode_bundle(bundle)
            vector_seq_concated = np.concatenate(vector_seqs, axis=0)
            rule.scale_balancer = CovarianceBasedScaleBalancer.from_feature_seqs(
                vector_seq_concated, dims
            )
        return rule

    @classmethod
    def create_default(
        cls,
        project_path: Path,
        include_image_encoder: bool = True,
        use_balancer: bool = True,
    ) -> "EncodingRule":

        cache_key = (project_path, include_image_encoder, use_balancer)
        if cache_key in _default_encoding_rule_cache:
            return _default_encoding_rule_cache[cache_key]

        bundle = EpisodeBundle.load(project_path)
        bundle_spec = bundle.spec

        encoders: List[EncoderBase] = []

        if include_image_encoder:
            image_encoder = ImageEncoder.create_default(project_path)
            encoders.append(image_encoder)

        if AngleVector in bundle_spec.type_shape_table:
            av_dim = bundle_spec.type_shape_table[AngleVector][0]
            av_idendical_encoder = VectorIdenticalEncoder.create(AngleVector, av_dim)
            encoders.append(av_idendical_encoder)

        if GripperState in bundle_spec.type_shape_table:
            gs_identital_func = VectorIdenticalEncoder.create(
                GripperState, bundle_spec.type_shape_table[GripperState][0]
            )
            encoders.append(gs_identital_func)

        if AnotherGripperState in bundle_spec.type_shape_table:
            ags_identital_func = VectorIdenticalEncoder.create(
                AnotherGripperState, bundle_spec.type_shape_table[AnotherGripperState][0]
            )
            encoders.append(ags_identital_func)

        tf_identical_func = VectorIdenticalEncoder.create(TerminateFlag, 1)
        encoders.append(tf_identical_func)

        p = CovarianceBasedScaleBalancer.get_json_file_path(project_path)
        if p.exists():  # use cached balacner
            balancer: Optional[CovarianceBasedScaleBalancer]
            if use_balancer:
                logger.warning(
                    "warn: loading cached CovarianceBasedScaleBalancer. This feature is experimental."
                )
                balancer = CovarianceBasedScaleBalancer.load(project_path)
            else:
                balancer = None
            encoding_rule = EncodingRule.from_encoders(
                encoders, bundle=None, scale_balancer=balancer
            )
        else:
            bundle_for_balancer: Optional[EpisodeBundle]
            if use_balancer:
                bundle_for_balancer = bundle
            else:
                bundle_for_balancer = None
            encoding_rule = EncodingRule.from_encoders(
                encoders, bundle=bundle_for_balancer, scale_balancer=None
            )

        # TODO: Move The following check to unittest? but it's diffcult becaues
        # using this function pre-require the existence of trained AE ...
        # so temporary, check using assertion

        # check if ordered propery. Keeping this order is important to satisfy the
        # backward-compatibility of this function.
        order_definition: List[Type[ElementBase]] = [
            RGBImage,
            DepthImage,
            RGBDImage,
            AngleVector,
            GripperState,
            AnotherGripperState,
            TerminateFlag,
        ]
        indices = [order_definition.index(key) for key in encoding_rule.keys()]
        is_ordered_propery = sorted(indices) == indices
        assert is_ordered_propery

        # update cache
        _default_encoding_rule_cache[cache_key] = encoding_rule

        return encoding_rule


@dataclass
class CompositeEncodingRule(EncodingRuleBase):
    rules: List[EncodingRule]

    def __post_init__(self):
        key_set_list = [set(rule.keys()) for rule in self.rules]
        set_entire = set.union(*key_set_list)
        no_intersection = len(set_entire) == sum([len(rule.keys()) for rule in self.rules])
        assert no_intersection

    def apply(self, edict: ElementDict) -> np.ndarray:
        vec = np.hstack([rule.apply(edict) for rule in self.rules])
        return vec

    def inverse_apply(self, vector: np.ndarray) -> ElementDict:
        head = 0
        elems = []
        for rule in self.rules:
            tail = head + rule.dimension
            edict = rule.inverse_apply(vector[head:tail])
            elems.extend(list(edict.values()))
            head = tail
        edict_merged = ElementDict(elems)
        return edict_merged

    @property
    def dimension_list(self) -> List[int]:
        dims = []
        for rule in self.rules:
            for encoder in rule.values():
                dims.append(encoder.output_size)
        return dims

    def __getitem__(self, key: Type[ElementBase]) -> EncoderBase:
        for rule in self.rules:
            if key in rule:
                return rule[key]
        raise KeyError

    def __iter__(self) -> Iterator[Type[ElementBase]]:
        return chain(*[rule.__iter__() for rule in self.rules])

    def __len__(self) -> int:
        return sum([len(rule) for rule in self.rules])



import argparse
from pathlib import Path

from mohou.file import get_project_path
from mohou.propagator import PropagatorBase, PropagatorSelection
from mohou.script_utils import visualize_lstm_propagation
from mohou.setting import setting
from mohou.types import Type

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-pn", type=str, default=setting.primary_project_name, help="project name")
    parser.add_argument("-pp", type=str, help="project path. preferred over pn.")
    parser.add_argument("-n", type=int, default=150, help="number of visualization")
    parser.add_argument("-model", type=str, default="lstm", help="select model")

    args = parser.parse_args()
    project_name: str = args.pn
    project_path_str: str = args.pp
    n_prop: int = args.n
    model_str: str = args.model

    if project_path_str is None:
        project_path = get_project_path(project_name)
    else:
        project_path = Path(project_path_str)

    prop_type: Type[PropagatorBase] = PropagatorSelection[model_str].value
    propagator = prop_type.create_default(project_path)
    visualize_lstm_propagation(project_path, propagator, n_prop)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Generic, List, Optional, Tuple, Type, TypeVar

import numpy as np
import torch

from mohou.constant import CONTINUE_FLAG_VALUE
from mohou.encoder import ImageEncoder
from mohou.encoding_rule import EncodingRule, EncodingRuleBase
from mohou.model import LSTM, PBLSTM
from mohou.model.common import ModelT
from mohou.model.experimental import (
    DisentangleLSTM,
    MarkovModelT,
    MarkovPredictionModel,
    ProportionalModel,
)
from mohou.trainer import TrainCache
from mohou.types import ElementDict, RGBImage, TerminateFlag
from mohou.utils import detect_device


@dataclass
class TerminateChecker:
    n_check_flag: int = 3
    terminate_threshold: float = 0.95

    def __call__(self, flags_predicted: List[TerminateFlag]) -> bool:
        # checking only latest terminate flags is not sufficient because
        # sometimes terminate flag value experiences sudden surge to 1 and
        # immediently goes back 0.
        if len(flags_predicted) < self.n_check_flag:
            return False

        flag_arr = np.array([flag.numpy().item() for flag in flags_predicted])
        return bool(np.all(flag_arr[-self.n_check_flag :] > self.terminate_threshold).item())


PropagatorT = TypeVar("PropagatorT", bound="PropagatorBase")


class PropagatorBase(ABC, Generic[ModelT]):
    propagator_model: ModelT
    encoding_rule: EncodingRuleBase

    @classmethod
    @abstractmethod
    def create_default(cls: Type[PropagatorT], porject_path: Path) -> PropagatorT:
        pass

    @property
    @abstractmethod
    def require_static_context(self) -> bool:
        pass

    @abstractmethod
    def feed(self, elem_dict: ElementDict) -> None:
        pass

    @abstractmethod
    def predict(self, n_prop: int) -> List[ElementDict]:
        pass

    @abstractmethod
    def reset(self) -> None:
        pass


class LSTMPropagatorBase(PropagatorBase[ModelT]):
    fed_state_list: List[np.ndarray]  # eatch state is equipped with flag
    n_init_duplicate: int
    is_initialized: bool
    predicted_terminated_flags: List[TerminateFlag]
    static_context: np.ndarray
    prop_hidden: bool = False
    _hidden: Optional[torch.Tensor] = None  # hidden state of lstm

    def __init__(
        self,
        lstm: ModelT,
        encoding_rule: EncodingRuleBase,
        n_init_duplicate: int = 0,
        prop_hidden: bool = False,
        device: Optional[torch.device] = None,
    ):

        if device is None:
            device = detect_device()
        encoding_rule.set_device(device)
        lstm.put_on_device(device)

        self.propagator_model = lstm
        self.encoding_rule = encoding_rule
        self.fed_state_list = []
        self.n_init_duplicate = n_init_duplicate
        self.prop_hidden = prop_hidden

        self.is_initialized = False
        self.predicted_terminated_flags = []

        if not self.require_static_context:  # auto set
            self.static_context = np.empty((0,))

    def is_terminated(self) -> bool:
        checker = TerminateChecker()
        return checker(self.predicted_terminated_flags)

    def reset(self) -> None:
        self.fed_state_list = []
        self.is_initialized = False
        self.predicted_terminated_flags = []

    @property
    def require_static_context(self) -> bool:
        if hasattr(self.propagator_model.config, "n_static_context"):
            return self.propagator_model.config.n_static_context > 0
        else:
            return False

    def set_static_context(self, value: np.ndarray) -> None:
        assert hasattr(self.propagator_model.config, "n_static_context")
        assert value.ndim == 1
        assert self.propagator_model.config.n_static_context == len(value)
        self.static_context = value

    def feed(self, elem_dict: ElementDict) -> None:
        self._feed(elem_dict)
        if not self.is_initialized:
            for _ in range(self.n_init_duplicate):
                self._feed(elem_dict)

        if not self.is_initialized:
            self.is_initialized = True

    def _feed(self, elem_dict: ElementDict) -> None:
        if TerminateFlag not in elem_dict:
            elem_dict[TerminateFlag] = TerminateFlag.from_bool(False)
        state_with_flag = self.encoding_rule.apply(elem_dict)
        self.fed_state_list.append(state_with_flag)

    def predict(self, n_prop: int, term_threshold: Optional[float] = None) -> List[ElementDict]:
        # prediction
        pred_state_list = self._predict(n_prop)
        elem_dict_list = []
        for pred_state in pred_state_list:
            elem_dict = self.encoding_rule.inverse_apply(pred_state)
            elem_dict_list.append(elem_dict)

        # update predicted_terminated_flags
        self.predicted_terminated_flags.append(elem_dict_list[0][TerminateFlag])

        return elem_dict_list

    def _predict(
        self, n_prop: int, force_continue: bool = False, term_threshold: Optional[float] = None
    ) -> List[np.ndarray]:
        pred_state_list: List[np.ndarray] = []

        assert self.static_context is not None, "forgot setting static_context ??"

        for i in range(n_prop):
            states = np.vstack(self.fed_state_list + pred_state_list)
            out, hidden = self._forward(states)

            if self.prop_hidden:
                # From the definition, propagating also hidden with the state is supporsed to
                # yield better prediction. But, from my experiment, it seems that performance
                # is actually slight worse than not doing it.
                # But need more experiment.
                self._hidden = hidden

            state_pred_torch = out[0, -1, :]
            state_pred = state_pred_torch.cpu().detach().numpy()

            if term_threshold is None:
                satisfy_break_condition = False
            else:
                satisfy_break_condition = state_pred[-1].item() > term_threshold

            if force_continue:
                state_pred[-1] = CONTINUE_FLAG_VALUE
            pred_state_list.append(state_pred)

            if satisfy_break_condition:
                break

        return pred_state_list

    def get_device(self) -> Optional[torch.device]:
        rule_device = self.encoding_rule.get_device()
        lstm_device = self.propagator_model.device

        if rule_device is None:
            return lstm_device
        else:
            assert rule_device == lstm_device
            return rule_device

    def set_device(self, device: torch.device) -> None:
        self.encoding_rule.set_device(device)
        self.propagator_model.put_on_device(device)

    @abstractmethod
    def _forward(self, states: np.ndarray) -> Tuple[torch.Tensor, torch.Tensor]:
        pass


PropagatorBaseT = TypeVar("PropagatorBaseT", bound=LSTMPropagatorBase)


class _LSTMPropagator(LSTMPropagatorBase[LSTM]):
    def _forward(self, states: np.ndarray) -> Tuple[torch.Tensor, torch.Tensor]:
        def numpy_to_unsqueezed_torch(arr: np.ndarray) -> torch.Tensor:
            device = self.get_device()
            return torch.from_numpy(arr).float().unsqueeze(dim=0).to(device)

        context_torch = numpy_to_unsqueezed_torch(self.static_context)
        states_torch = numpy_to_unsqueezed_torch(states)
        out, hidden = self.propagator_model.forward(states_torch, context_torch, self._hidden)
        return out, hidden


class LSTMPropagator(_LSTMPropagator):
    @classmethod
    def create_default(cls, project_path: Path) -> "LSTMPropagator":
        tcach_lstm = TrainCache.load(project_path, LSTM)
        encoding_rule = EncodingRule.create_default(project_path)
        return cls(tcach_lstm.best_model, encoding_rule)


class ChimeraPropagator(_LSTMPropagator):
    @classmethod
    def create_default(cls, project_path: Path) -> "ChimeraPropagator":
        from mohou.model.chimera import Chimera

        tcache_chimera = TrainCache.load(project_path, Chimera)
        chimera_model = tcache_chimera.best_model

        rule = EncodingRule.create_default(project_path)
        rule[RGBImage] = ImageEncoder.from_auto_encoder(chimera_model.ae)
        return cls(chimera_model.lstm, rule)


class PBLSTMPropagator(LSTMPropagatorBase[PBLSTM]):
    parametric_bias: np.ndarray

    @classmethod
    def create_default(cls, project_path: Path) -> "PBLSTMPropagator":
        tcach_lstm = TrainCache.load(project_path, PBLSTM)
        encoding_rule = EncodingRule.create_default(project_path)
        prop = cls(tcach_lstm.best_model, encoding_rule)
        prop.set_pb_to_zero()
        return prop

    def set_parametric_bias(self, value: np.ndarray) -> None:
        assert value.ndim == 1
        assert len(value) == self.propagator_model.config.n_pb_dim
        self.parametric_bias = value

    def set_pb_to_zero(self) -> None:
        n_pb_dim = self.propagator_model.config.n_pb_dim
        vec = np.zeros(n_pb_dim)
        self.set_parametric_bias(vec)

    def _forward(self, states: np.ndarray) -> Tuple[torch.Tensor, torch.Tensor]:
        def numpy_to_unsqueezed_torch(arr: np.ndarray) -> torch.Tensor:
            device = self.get_device()
            return torch.from_numpy(arr).float().unsqueeze(dim=0).to(device)

        context_torch = numpy_to_unsqueezed_torch(self.static_context)
        pb_torch = numpy_to_unsqueezed_torch(self.parametric_bias)
        states_torch = numpy_to_unsqueezed_torch(states)
        out, hidden = self.propagator_model.forward(
            states_torch, pb_torch, context_torch, self._hidden
        )
        return out, hidden


class DisentangleLSTMPropagator(LSTMPropagatorBase[DisentangleLSTM]):
    @classmethod
    def create_default(cls, project_path: Path) -> "DisentangleLSTMPropagator":
        tcache = TrainCache.load(project_path, DisentangleLSTM)
        encoding_rule = EncodingRule.create_default(project_path)
        return cls(tcache.best_model, encoding_rule)

    def _forward(self, states: np.ndarray) -> Tuple[torch.Tensor, torch.Tensor]:
        def numpy_to_unsqueezed_torch(arr: np.ndarray) -> torch.Tensor:
            device = self.get_device()
            return torch.from_numpy(arr).float().unsqueeze(dim=0).to(device)

        states_torch = numpy_to_unsqueezed_torch(states)
        pred_torch = self.propagator_model.forward(states_torch)
        hidden = torch.empty((0,)).to(self.get_device())  # dummy
        return pred_torch, hidden


MarkovPropagatorT = TypeVar("MarkovPropagatorT", bound="MarkovPropagator")


@dataclass
class MarkovPropagator(PropagatorBase[MarkovModelT]):
    model: MarkovModelT
    encoding_rule: EncodingRuleBase
    fed_vector: Optional[np.ndarray] = None

    @classmethod
    @abstractmethod
    def get_model_type(cls) -> Type[MarkovModelT]:
        pass

    @classmethod
    def create_default(cls: Type[MarkovPropagatorT], project_path: Path) -> MarkovPropagatorT:
        model_type = cls.get_model_type()
        tcache = TrainCache.load(project_path, model_type)
        encoding_rule = EncodingRule.create_default(project_path)
        return cls(tcache.best_model, encoding_rule)

    @property
    def require_static_context(self) -> bool:
        # currently not support static context
        return False

    def feed(self, elem_dict: ElementDict) -> None:
        feature_vector = self.encoding_rule.apply(elem_dict)
        self.fed_vector = feature_vector

    def predict(self, n_prop: int) -> List[ElementDict]:
        assert self.fed_vector is not None
        feeding_vector_list = [self.fed_vector]
        for _ in range(n_prop):
            latest_vector = feeding_vector_list[-1]
            torch_vector = torch.from_numpy(latest_vector).float()
            seq_sample = torch_vector.unsqueeze(0).unsqueeze(0)
            pred_sample = self.model.forward(seq_sample).squeeze()
            feeding_vector_list.append(pred_sample.cpu().detach().numpy())
        pred_vector_list = feeding_vector_list[1:]
        pred_edict_list = [self.encoding_rule.inverse_apply(vec) for vec in pred_vector_list]
        return pred_edict_list

    def reset(self) -> None:
        pass


@dataclass
class ProportionalModelPropagator(MarkovPropagator[ProportionalModel]):
    @classmethod
    def get_model_type(cls) -> Type[ProportionalModel]:
        return ProportionalModel


@dataclass
class MarkovPredictionModelPropagator(MarkovPropagator[MarkovPredictionModel]):
    @classmethod
    def get_model_type(cls) -> Type[MarkovPredictionModel]:
        return MarkovPredictionModel


class PropagatorSelection(Enum):
    lstm = LSTMPropagator
    pblstm = PBLSTMPropagator
    chimera = ChimeraPropagator
    proportional = ProportionalModelPropagator
    markov_prediction = MarkovPredictionModelPropagator
    disentangle = DisentangleLSTMPropagator
